unit bindata;
(* ===========================================================================
 * bindata.dcu - tBinaryDataSet : A custom data set which uses a flat binary
 *             structured datafile for single client usage only.
 *
 * Author:  Steven R. Kamradt
 * Copyright (C) 1997, Steven R. Kamradt
 *
 * Many thanks to the following delphi programmers who have made this control
 * possible:
 *
 *         Marco Cantu [Author of "Delphi Developers Handbook"]
 *         Tony BenBrahim [http://www.neosoft.com/~startech/week1/customds1.htm]
 *         Daniel Gavarin [news:borland.public.delphi.vcl.components.writing]
 *         Peter Szymiczek [UNDU #8, QSort algorithm]
 *
 * Revision History:
 * version 0.9 - 01/19/98 : Initial version, merged Danie Gavarin's code with
 *                          the Marco's code from tDdhObjectDataSet.
 *                        : Modified header to support a record deletion
 *                          chain. It does not reuse deleted records, yet. :)
 *               01/20/98 : Corrected bug which disallowed the use of tDbCheckbox
 *                          components.
 *                        : Added PackTable method to remove deleted records
 *                          from the dataset datafile.
 *               02/03/98 : ReadOnly property now available which will open
 *                          the table in readonly mode.
 *               02/26/98 : Added SortTable method to sort the table. Also
 *                          added an UnSortTable method to randomly distribute
 *                          records to test/benchmark the sorting algorithms.
 *               02/27/98 : Changed SortTable method to use QuickSort sorting
 *                          algorithm for increased performance.
 *               03/09/98 : Added documentation file BINDATA.TXT and prepared
 *                          for public release.
 * version 0.91- 04/20/98 : Thanks to Shin Yang, and Tom White who pointed out a 
 *                          bug in the CreateTable method.  A working fix has been 
 *                          made, however it is important to note that when 
 *                          dynamicaly creating tables, you MUST use a fresh 
 *                          instance of tBinaryDataSet.  The reasons for this are 
 *                          not clear, but it seems to be a bug in the tDataSet 
 *                          object. See the example in Exmain.pas.
 * ===========================================================================
 *)

interface

uses
  SysUtils, Classes, Db;

type
  EBinaryDataSetError = class (Exception);

  PRecInfo = ^TRecInfo;
  TRecInfo = record
    Bookmark: Longint;
    BookmarkFlag: TBookmarkFlag;
  end;

type
  tDataFileHeader = record
    GenHeader : Array[1..4] of Char;
    VersionNumber : Integer;
    RecordSize : Integer;
    RecordCount : Integer;
    NumberFields : Integer;
    FirstDeleted : Integer;
    LastDeleted : Integer;
    DeletedCount : Integer;
    StartData : Integer;
  end;

Type
  pRecordHeader = ^tRecordHeader;
  tRecordHeader = record
    DeletedFlag : Integer;
  end;

type
  TBinaryDataSet = class(TDataSet)
  protected
    FStream: TStream; // the physical table
    FTableName: string; // table path and file name
    fDataFileHeader : tDataFileHeader;
    // record data
    fRecordHeaderSize : Integer; // The size of the record header
    FRecordCount, // current number of record
    FRecordSize, // the size of the actual data
    FRecordBufferSize, // data + housekeeping (TRecInfo)
    FRecordInfoOffset, // offset of RecInfo in record buffer
    FCurrentRecord, // current record (0 to FRecordCount - 1)
    BofCrack, // before the first record (crack)
    EofCrack: Integer; // after the last record (crack)
    // file header size
    FDataFileHeaderSize : Integer;
    // status
    FIsTableOpen: Boolean;
    // field offsets in record
    FFieldOffset: TList;
    //Enhancements
    fReadOnly : Boolean;
  protected
    // TDataSet virtual abstract method
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar;
      Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar):
      TBookmarkFlag; override;
    function GetFieldData(Field: TField;
      Buffer: Pointer): Boolean; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode;
      DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer;
      Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(
      Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar;
      Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar;
      Data: Pointer); override;
    procedure SetFieldData(Field: TField;
      Buffer: Pointer); override;

    // TDataSet virtual method (optional)
    function GetRecordCount: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    function GetRecNo: Integer; override;
    Procedure WriteHeader;
  private
    Procedure _ReadRecord(Buffer:PChar;IntRecNum:Integer);
    Procedure _WriteRecord(Buffer:PChar;IntRecNum:Integer);
    Procedure _AppendRecord(Buffer:PChar);
    Procedure _SwapRecords(Rec1,REc2:Integer);
    Function _CompareRecords(SortFields:Array of String;Rec1,Rec2:Integer):Integer;
  public
    constructor Create(AOwner:tComponent); override;
    procedure CreateTable;
    Procedure PackTable;
    Procedure SortTable(SortFields : Array of String);
    Procedure UnsortTable;
  published
    property TableName: string
      read FTableName write FTableName;
    property ReadOnly : Boolean read fReadOnly write fReadonly default False;
    property DataFileHeader : tDataFileHeader read fDataFileHeader;
    // redeclared data set properties
    property Active;
    property Filtered;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;

procedure Register;

implementation

uses
  TypInfo, Dialogs, Windows, Forms, Controls, IniFiles;

Const
  dfhVersionNumber = 13;

TYPE
  PBufArray = ^BufArray;
  BufArray = Array[0..0] of Char;

type
  PBDescribField = ^TBDescribField;
  TBDescribField = packed Record
    Name : Array[0..27] of Char;
    DataType : TFieldType;
    DataSize : Integer;
    Size : Integer;
    Precision : Integer;
  end;


// ****************************************************************************
// Low Level Routines for accessing an internal record

// ____________________________________________________________________________
// tBinaryDataSet._ReadRecord
Procedure tBinaryDataSet._ReadRecord(Buffer:PChar;IntRecNum:Integer);
  {-Read a record based on the internal record number (absolute)}
BEGIN
  FStream.Position := FDataFileHeader.StartData + (FRecordSize * IntRecNum);
  FStream.ReadBuffer (Buffer^, FRecordSize);
END;

// ____________________________________________________________________________
// tBinaryDataSet._WriteRecord
Procedure tBinaryDataSet._WriteRecord(Buffer:PChar;IntRecNum:Integer);
  {-Write a record based on the internal record number (absolute)}
BEGIN
  FStream.Position := FDataFileHeader.StartData + (FRecordSize * IntRecNum);
  FStream.WriteBuffer (Buffer^, FRecordSize);
END;

// ____________________________________________________________________________
// tBinaryDataSet._AppendRecord
Procedure tBinaryDataSet._AppendRecord(Buffer:PChar);
BEGIN
  FStream.Seek (0, soFromEnd);
  FStream.WriteBuffer (Buffer^, FRecordSize);
END;

/////////////////////////////////////////////////
////// Part I:
////// Initialization, opening, and closing
/////////////////////////////////////////////////

// ____________________________________________________________________________
// tBinaryDataSet.InternalOpen
// I: open the table/file
procedure tBinaryDataSet.InternalOpen;
begin
  // check if the file exists
  if not FileExists (FTableName) then
    raise eBinaryDataSetError.Create ('Open: Table file not found');

  // create a stream for the file
  if fReadOnly then
    fStream := tFileStream.Create( fTableName, fmOpenRead + fmShareDenyWrite)
  else
    FStream := TFileStream.Create (FTableName, fmOpenReadWrite + fmShareExclusive);
  fStream.ReadBuffer(fDataFileHeader,FDataFileHeaderSize);
  if fDataFileHeader.GenHeader <> 'BDS'+^Z then
    Raise EBinaryDataSetError.Create('Not a binary data set file.');
  if fDataFileHeader.VersionNumber <> dfhVersionNumber then
    Raise EBinaryDataSetError.Create('Invalid data file version.');

  // initialize the field definitions
  // (another virtual abstract method of TDataSet)
  InternalInitFieldDefs;

  // if there are no persistent field objects,
  // create the fields dynamically
  if DefaultFields then
    CreateFields;
  // connect the TField objects with the actual fields
  BindFields (True);

  // get the number of records and check size
  fRecordCount := fDataFileHeader.RecordCount;
  if (fRecordCount > 0) and
    (fDataFileHeader.RecordSize <> fRecordSize) then
      Raise eBinaryDataSetError.Create('Internal open: File record size mismatch.');

  if (fDataFileHeader.StartData + (fRecordCount + fDataFileHeader.DeletedCount) * fRecordSize) <> fStream.Size then
    Raise eBinaryDataSetError.Create('Internal open: Invalid record size.'+^M+
                                     'StartData   : '+IntToSTr(fDataFileHeader.StartData)+^M+
                                     'fRecordCount: '+IntToStr(fRecordCount)+^M+
                                     'fRecordSize : '+IntToStr(fRecordSize)+^M+
                                     'fStream.SIze: '+IntToStr(fStream.Size));

  // sets cracks and record position
  BofCrack := -1;
  EofCrack := FRecordCount;
  FCurrentRecord := BofCrack;

  FRecordInfoOffset := FRecordSize;
  FRecordBufferSize := FRecordSize + sizeof (TRecInfo);

  // set the bookmark size
  BookmarkSize := sizeOf (Integer);

  // everything OK: table is now open
  FIsTableOpen := True;

  // ShowMessage ('InternalOpen: RecCount: ' + IntToStr (FRecordCount));
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalInitFieldDefs
// I: define the fields
procedure tBinaryDataSet.InternalInitFieldDefs;
var
  Il : Integer;
  DescribF : TBDescribField;
  TmpFieldOffset : Integer;
  IX : Integer;
begin
  FFieldOffset := TList.Create;
  TRY
    for IX := FieldDefs.Count-1 downto 0 do
      if Assigned(FieldDefs.Items[IX]) then
        FieldDefs.Items[IX].Free;
  EXCEPT
    // GPF HERE!
    //  FieldDefs.Clear {-GENERATES A GPF!}
  END;
  FStream.Seek(fDataFileHeaderSize,soFromBeginning);
  Assert(fDataFileHeader.NumberFields>0,'Invalid number of fields.');

  TmpFieldOffset := fRecordHeaderSize;
  if (fDataFileHeader.NumberFields>0) then
  begin
    for Il:=0 to fDataFileHeader.NumberFields-1 do
    begin
      FStream.Read(DescribF,SizeOf(DescribF));
      TFieldDef.Create(FieldDefs, DescribF.Name,DescribF.DataType,
        DescribF.Size,False,Il+1);
      FFieldOffset.Add (Pointer (TmpFieldOffset));
      Inc(tmpFieldOffset,DescribF.DataSize);
    end;
    fRecordSize := tmpFieldOffset;
  end;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalClose
// I: close the table/file
procedure tBinaryDataSet.InternalClose;
begin
  // if required, save updated header
  if (fDataFileHeader.RecordCount <> fRecordCount) or
    (fDataFileHeader.RecordSize = 0) then
    BEGIN
      fDataFileHeader.RecordSize := fRecordSize;
      fDataFileHeader.RecordCount := fRecordCount;
      WriteHeader;
    END;
  // free the internal list field offsets
  if Assigned (FFieldOffset) then
    FFieldOffset.Free;

  // disconnet field objects
  BindFields (False);
  // destroy field object (if not persistent)
  if DefaultFields then
    DestroyFields;

  // close the file
  FIsTableOpen := False;
  FStream.Free;
end;

// ____________________________________________________________________________
// tBinaryDataSet.IsCursorOpen
// I: is table open
function tBinaryDataSet.IsCursorOpen: Boolean;
begin
  Result := FIsTableOpen;
end;

// ____________________________________________________________________________
// tBinaryDataSet.WriteHeader
procedure tBinaryDataSet.WriteHeader;
begin
  Assert(FStream<>nil,'fStream=Nil');
  FSTream.Seek(0,soFromBeginning);
  FStream.WriteBuffer(fDataFileHeader,fDataFileHeaderSize);
end;

// ____________________________________________________________________________
// tBinaryDataSet.Create
constructor tBinaryDataSet.Create(AOwner:tComponent);
BEGIN
  inherited create(aOwner);
  fDataFileHeaderSize := SizeOf(tDataFileHeader);
  fRecordHeaderSize := SizeOf(tRecordHeader);
END;

// ____________________________________________________________________________
// tBinaryDataSet.CreateTable
// I: Create a new table/file
procedure tBinaryDataSet.CreateTable;
var
  Ix : Integer;
  DescribF : TBDescribField;
  Offs : Integer;
begin
  CheckInactive;
  //  InternalInitFieldDefs;
  // create the new file
  if FileExists (FTableName) and
    (MessageDlg ('File ' + FTableName +
      ' already exists. OK to override?',
      mtConfirmation, [mbYes, mbNo], 0) = mrNo) then
    Exit;
  if FieldDefs.Count = 0 then
  begin
    for Ix := 0 to FieldCount - 1 do
    begin
      with Fields[Ix] do
      begin
        if FieldKind = fkData then
          FieldDefs.Add(FieldName,DataType,Size,Required);
      end;
    end;
  end;
  FStream := TFileStream.Create (FTableName,
    fmCreate or fmShareExclusive);
  try
    FillChar(fDataFileHeader,fDataFileHeaderSize,0);
    fDataFileHeader.GenHeader := 'BDS'+^Z;
    fDataFileHeader.VersionNumber := dfhVersionNumber;
    fDataFileHeader.RecordSize := 0; // filled later
    fDataFileheader.RecordCount := 0; // empty
    fDataFileHeader.NumberFields := FieldDefs.Count;
    fDataFileHeader.LastDeleted := -1;
    fDataFileHeader.FirstDeleted := -1;
    fDataFileHeader.DeletedCount := 0;
    fDataFileHeader.StartData := -1;
    WriteHeader;
    Offs:=fRecordHeaderSize;
    for Ix:=0 to FieldDefs.Count-1 do
      begin
        with FieldDefs.Items[Ix] do
        begin
          FillChar(DescribF,SizeOf(DescribF),#0);
          DescribF.DataType:=DataType;
          StrCopy(DescribF.Name,PChar(FieldDefs.Items[Ix].Name));
          DescribF.Size:=Size;
          DescribF.Precision:=Precision;
          case DataType of
            ftString:
              DescribF.DataSize := Size+SizeOf(INteger);
            ftBoolean, ftSmallInt, ftWord:
              DescribF.DataSize := 2;
            ftInteger, ftDate, ftTime:
              DescribF.DataSize := 4;
            ftFloat, ftCurrency, ftDateTime:
              //DescribF.DataSize:= Fields[Ix].DataSize; //8;
              DescribF.DataSize := 8;
          else
            raise EBinaryDataSetError.Create (
               'InitFieldsDefs: Unsupported field type');
          end;
          Inc(Offs,DescribF.DataSIze);
          //INC(fDataFileHeader.NumberFields);
          FStream.Write(DescribF,SizeOf(DescribF));
        end;
      end;
    fDataFileHeader.StartData := FStream.Position;
    fDataFileHeader.RecordSize := Offs;
    FRecordSize := Offs;
    WriteHeader;
  finally
    // close the file
    fStream.Free;
  end;
end;

// ____________________________________________________________________________
// tBinaryDataSet.PackTable
//Enhancement: Remove all deleted items from the table.
Procedure tBinaryDataSet.PackTable;
var
  NewStream, OldStream : tStream;
  PC : PChar;
  Ix : Integer;
  DescribF : TBDescribField;
  NewDataFileHeader : tDataFileHeader;
  DataBuffer : Pointer;
BEGIN
  OldStream := Nil;
  NewStream := Nil;
  CheckInactive;
//  if Active then
//    raise eBinaryDataSetError.Create ('Dataset must be closed before packing.');
  if fTableName = '' then
    raise EBinaryDataSetError.Create('Table name not specified.');
  if not FileExists (FTableName) then
    raise EBinaryDataSetError.Create('Table '+fTableName+' does not exist.');
  PC := @fTablename[1];
  CopyFile(PChar(PC),PChar(PC+',old'+#0),False);
  // create the new file
  if FieldDefs.Count = 0 then
  begin
    for Ix := 0 to FieldCount - 1 do
    begin
      with Fields[Ix] do
      begin
        if FieldKind = fkData then
          FieldDefs.Add(FieldName,DataType,Size,Required);
      end;
    end;
  end;
  TRY
    NewStream := TFileStream.Create (FTableName+',new',
      fmCreate or fmShareExclusive);
    OldStream := tFileStream.Create (fTableName+',old',
      fmOpenRead or fmShareExclusive);
    OldStream.ReadBuffer(NewDataFileHeader,fDataFileHeaderSize);
    NewDataFileHeader.DeletedCount := 0;
    NewDataFileHeader.LastDeleted := -1;
    NewDataFileHeader.FirstDeleted := -1;
    NewStream.WriteBuffer(NewDataFileHeader,fDataFileHeaderSize);
    for IX := 0 to NewDataFileHeader.NumberFields-1 do
      BEGIN
        OldStream.Read(DescribF,SizeOf(DescribF));
        NewStream.Write(DescribF,SizeOf(DescribF));
      END;
    GetMem(DataBuffer,NewDataFileHeader.RecordSize);
    REPEAT
      IX := OldStream.Read(DataBuffer^,NewDataFileHeader.RecordSize);
      if (IX = NewDataFileHeader.REcordSize) and (pRecordHeader(DataBuffer)^.DeletedFlag = 0) then
        NewStream.WRite(DataBuffer^,NewDataFileHeader.RecordSize);
    Until IX <> NewDataFileHeader.RecordSize;
    FreeMem(DataBuffer,NewDataFileHeader.RecordSize);
  finally
    // close the file
    NewStream.Free;
    OldStream.Free;
  end;
  CopyFile(PChar(PC+',new'+#0),PChar(PC),False);
  DeleteFile(Pchar(PC+',new'+#0));
  DeleteFile(Pchar(PC+',old'+#0));
END;

// ____________________________________________________________________________
// tBinaryDataSet._SwapRecords
// Enhancement: Quick swap of two records.  Used primarily for sorting.
Procedure tBinaryDataSet._SwapRecords(Rec1,REc2:Integer);
VAR
  Buffer1, Buffer2 : PChar;
  Bookmark1, BOokmark2 : TBookmarkFlag;
BEGIN
  Rec1 := Rec1 - 1;
  Rec2 := Rec2 - 1;
  if Rec1 < 0 then Exit;
  if Rec2 < 0 then Exit;
  Buffer1 := AllocRecordBuffer;
  Buffer2 := AllocRecordBuffer;
  _ReadRecord(Buffer1,Rec1);
  _ReadRecord(Buffer2,Rec2);
  Bookmark1 := GetBookmarkFlag(Buffer1);
  Bookmark2 := GetBookmarkFlag(Buffer2);
  SetBookmarkFlag(Buffer1,Bookmark2);
  SetBookmarkFlag(Buffer2,Bookmark1);
  _WriteRecord(Buffer1,Rec2);
  _WriteRecord(Buffer2,Rec1);
  StrDispose(Buffer1);
  StrDispose(Buffer2);
END;

// ____________________________________________________________________________
// tBinaryDataSet._CompareRecords
// Compare two records.  Returns -1 if REC1 < REC2, 0 if REC1 = REC2, or
// 1 if REC1 > REC2.
Function tBinaryDataSet._CompareRecords(SortFields:Array of String;Rec1,Rec2:Integer):Integer; FAR;
{-Compare the records Rec1, Rec2 and return -1 if Rec1 < Rec2, 0 if Rec1 = Rec2,
  1 if Rec1 > Rec2 }
VAR
  IX : Integer;

  Function CompareHelper(KeyId:String;Rec1,Rec2:Integer):Integer;
  VAR
    SKey1, SKey2 : String;
    IKey1, IKey2 : Integer;
    fKey1, fKey2 : Double;
    dKey1, dKey2 : tDateTime;
    CompareType : tFieldType;
    KeyField : tField;
  BEGIN
    KeyField := FieldByName(KeyID);
    CompareType := KeyField.DataType;
    Case CompareType of
      ftFloat,
      ftCurrency,
      ftBCD :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          fKey1 := KeyField.AsFloat;
          _ReadRecord(ActiveBuffer,Rec2-1);
          fKey2 := KeyField.AsFloat;
          if fKey1 < fKey2 then
            Result := -1
          else
            if fKey1 > fKey2 then
              Result := 1
            else
              Result := 0;
        END;
      ftSmallInt,
      ftInteger,
      ftWord :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          IKey1 := KeyField.AsInteger;
          _ReadRecord(ActiveBuffer,Rec2-1);
          IKey2 := KeyField.AsInteger;
          if IKey1 < IKey2 then
            Result := -1
          else
            if IKey1 > IKey2 then
              Result := 1
            else
              Result := 0;
        END;
      ftDate,
      ftTime,
      ftDateTime :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          dKey1 := KeyField.AsDateTime;
          _ReadRecord(ActiveBuffer,Rec2-1);
          dKey2 := KeyField.AsDateTime;
          if dKey1 < dKey2 then
            Result := -1
          else
            if dKey1 > dKey2 then
              Result := 1
            else
              Result := 0;
        END;
      else
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          SKey1 := KeyField.AsString;
          _ReadRecord(ActiveBuffer,Rec2-1);
          SKey2 := KeyField.AsString;
          if SKey1 < SKey2 then
            Result := -1
          else
            if SKey1 > SKey2 then
              Result := 1
            else
              Result := 0;
        END;
    END;
  END;

BEGIN
  IX := 0;
  REPEAT // Loop through all available sortfields until not equal or no more sort fiels.
    Result := CompareHelper(SortFields[IX],Rec1,Rec2);
    Inc(IX);
  UNTIL (Result <> 0) or (IX > High(SortFields));
END;


// ____________________________________________________________________________
// tBinaryDataSet.SortTable
// Enhancement: Sort the table by the fields passed.
Procedure tBinaryDataSet.SortTable(SortFields : Array of String);

  { This is the main sorting routine. It is passed the number of elements and the
    two callback routines. The first routine is the function that will perform
    the comparison between two elements. The second routine is the procedure that
    will swap two elements if necessary } // Source: UNDU #8

  procedure QSort(uNElem: Integer);
  { uNElem - number of elements to sort }

    procedure qSortHelp(pivotP: Integer; nElem: word);
    label
      TailRecursion,
      qBreak;
    var
      leftP, rightP, pivotEnd, pivotTemp, leftTemp: word;
      lNum: Integer;
      retval: integer;
    begin
      TailRecursion:
        if (nElem <= 2) then
          begin
            if (nElem = 2) then

              begin
                rightP := pivotP +1;
                if (_CompareRecords(SortFields,pivotP, rightP) > 0) then
                  _SwapRecords(pivotP, rightP);
              end;
            exit;
          end;
        rightP := (nElem -1) + pivotP;
        leftP :=  (nElem shr 1) + pivotP;
        { sort pivot, left, and right elements for "median of 3" }
        if (_CompareRecords(SortFields,leftP, rightP) > 0) then _SwapRecords(leftP, rightP);
        if (_CompareRecords(SortFields,leftP, pivotP) > 0) then _SwapRecords(leftP, pivotP)

        else if (_CompareRecords(SortFields,pivotP, rightP) > 0) then _SwapRecords(pivotP, rightP);
        if (nElem = 3) then
          begin
            _SwapRecords(pivotP, leftP);
            exit;
          end;
        { now for the classic Horae algorithm }
        pivotEnd := pivotP + 1;
        leftP := pivotEnd;
        repeat
          retval := _CompareRecords(SortFields,leftP, pivotP);
          while (retval <= 0) do
            begin
              if (retval = 0) then
                begin
                  _SwapRecords(LeftP, PivotEnd);
                  Inc(PivotEnd);
                end;
              if (leftP < rightP) then
                Inc(leftP)
              else
                goto qBreak;
              retval := _CompareRecords(SortFields,leftP, pivotP);
            end; {while}
          while (leftP < rightP) do
            begin

              retval := _CompareRecords(SortFields,pivotP, rightP);
              if (retval < 0) then
                Dec(rightP)
              else
                begin
                  _SwapRecords(leftP, rightP);
                  if (retval <> 0) then
                    begin
                      Inc(leftP);
                      Dec(rightP);
                    end;
                  break;
                end;
            end; {while}

        until (leftP >= rightP);
      qBreak:
        if (_CompareRecords(SortFields,leftP, pivotP) <= 0) then Inc(leftP);
        leftTemp := leftP -1;
        pivotTemp := pivotP;
        while ((pivotTemp < pivotEnd) and (leftTemp >= pivotEnd)) do
          begin
            _SwapRecords(pivotTemp, leftTemp);
            Inc(pivotTemp);
            Dec(leftTemp);
          end; {while}
        lNum := (leftP - pivotEnd);
        nElem := ((nElem + pivotP) -leftP);

        if (nElem < lNum) then
          begin
            qSortHelp(leftP, nElem);
            nElem := lNum;
          end
        else
          begin
            qSortHelp(pivotP, lNum);
            pivotP := leftP;
          end;
        goto TailRecursion;
      end; {qSortHelp }

  begin
    if (uNElem < 2) then  exit; { nothing to sort }
    qSortHelp(1, uNElem);
  end; { QSort }


BEGIN
  CheckActive;
  if fReadOnly then
    raise eBinaryDataSetError.Create ('Dataset must be opened for read/write to perform sort.');
//  if fDataFileHeader.DeletedCount > 0 then
//    BEGIN
//      Close;
//      PackTable;
//      Open;
//    END;
  QSort(FRecordCount + fDataFileHeader.DeletedCount);
  First;
END;

// ____________________________________________________________________________
// tBinaryDataSet.UnsortTable
// Used to help test the sort routine.  Attempts to generate a random
// dispersment of the records in the dataset.
Procedure tBinaryDataSet.UnsortTable;
Var
  IX : Integer;
BEGIN
  First;
  Randomize;
  for IX := 0 to RecordCOunt do
    BEGIN
      _SwapRecords(IX,Random(RecordCount+1));
    END;
  First;
END;

////////////////////////////////////////
////// Part II:
////// Bookmarks management and movement
////////////////////////////////////////

// ____________________________________________________________________________
// tBinaryDataSet.InternalGotoBookmark
// II: set the requested bookmark as current record
procedure tBinaryDataSet.InternalGotoBookmark (Bookmark: Pointer);
var
  ReqBookmark: Integer;
begin
  ReqBookmark := PInteger (Bookmark)^;
//  ShowMessage ('InternalGotoBookmark: ' +
//    IntToStr (ReqBookmark));
  if (ReqBookmark >= 0) and (ReqBookmark < FRecordCount + fDataFileHeader.DeletedCount) then
    FCurrentRecord := ReqBookmark
  else
    raise eBinaryDataSetError.Create ('Bookmark ' +
      IntToStr (ReqBookmark) + ' not found');
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalSetToRecord
// II: same as above (but passes a buffer)
procedure tBinaryDataSet.InternalSetToRecord (Buffer: PChar);
var
  ReqBookmark: Integer;
begin
//  ShowMessage ('InternalSetToRecord');
  ReqBookmark := PRecInfo(Buffer + FRecordInfoOffset).Bookmark;
  InternalGotoBookmark (@ReqBookmark);
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetBookmarkFlag
// II: retrieve bookmarks flags from buffer
function tBinaryDataSet.GetBookmarkFlag (
  Buffer: PChar): TBookmarkFlag;
begin
//  ShowMessage ('GetBookmarkFlag');
  Result := PRecInfo(Buffer + FRecordInfoOffset).BookmarkFlag;
end;

// ____________________________________________________________________________
// tBinaryDataSet.SetBookmarkFlag
// II: change the bookmark flags in the buffer
procedure tBinaryDataSet.SetBookmarkFlag (Buffer: PChar;
  Value: TBookmarkFlag);
begin
//  ShowMessage ('SetBookmarkFlag');
  PRecInfo(Buffer + FRecordInfoOffset).BookmarkFlag := Value;
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetBookmarkData
// II: read the bookmark data from record buffer
procedure tBinaryDataSet.GetBookmarkData (
  Buffer: PChar; Data: Pointer);
begin
//  ShowMessage ('GetBookmarkData: ' + IntToStr (PRecInfo(Buffer + FRecordInfoOffset).Bookmark));
  PInteger(Data)^ :=
    PRecInfo(Buffer + FRecordInfoOffset).Bookmark;
end;

// ____________________________________________________________________________
// tBinaryDataSet.SetBookmarkData
// II: set the bookmark data in the buffer
procedure tBinaryDataSet.SetBookmarkData (
  Buffer: PChar; Data: Pointer);
begin
//  ShowMessage ('SetBookmarkData: ' + IntToStr (PInteger(Data)^));
  PRecInfo(Buffer + FRecordInfoOffset).Bookmark :=
    PInteger(Data)^;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalFirst
// II: Go to a special position before the first record
procedure tBinaryDataSet.InternalFirst;
begin
  FCurrentRecord := BofCrack;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalLast
// II: Go to a special position after the last record
procedure tBinaryDataSet.InternalLast;
begin
  EofCrack := FRecordCount + fDataFileHeader.DeletedCount;
  FCurrentRecord := EofCrack;
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetRecordCount
// II (optional): Record count
function tBinaryDataSet.GetRecordCount: Longint;
begin
  CheckActive;
  Result := FRecordCount;
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetRecNo
// II (optional): Get the number of the current record
function tBinaryDataSet.GetRecNo: Longint;
begin
  UpdateCursorPos;
  if FCurrentRecord < 0 then
    Result := 1
  else
    Result := FCurrentRecord + 1;
end;

// ____________________________________________________________________________
// tBinaryDataSet.SetRecNo
// II (optional): Move to the given record number
procedure tBinaryDataSet.SetRecNo(Value: Integer);
begin
  CheckBrowseMode;
  if (Value > 1) and (Value <= FRecordCount) then
  begin
    FCurrentRecord := Value - 1;
    Resync([]);
  end;
end;

//////////////////////////////////////////
////// Part III:
////// Record buffers and field management
//////////////////////////////////////////

// ____________________________________________________________________________
// tBinaryDataSet.GetRecordSize
/// III: Determine the size of each record buffer in memory
function tBinaryDataSet.GetRecordSize: Word;
begin
  Result := FRecordSize; // data only
end;

// ____________________________________________________________________________
// tBinaryDataSet.AllocRecordBuffer
/// III: Allocate a buffer for the record
function tBinaryDataSet.AllocRecordBuffer: PChar;
begin
  Result := StrAlloc(FRecordBufferSize);
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalInitRecord
// III: Initialize the record (set to zero)
procedure tBinaryDataSet.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, FRecordBufferSize, 0);
end;

// ____________________________________________________________________________
// tBinaryDataSet.FreeRecordBuffer
// III: Free the buffer
procedure tBinaryDataSet.FreeRecordBuffer (var Buffer: PChar);
begin
  StrDispose(Buffer);
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetRecord
// III: Retrieve data for current, previous, or next record
// (eventually moving to it) and return the status
function tBinaryDataSet.GetRecord(Buffer: PChar;
  GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  Acceptable : Boolean;
begin
  result := grOk;
  Acceptable := False;
  if FRecordCount < 1 then
    Result := grEOF
  else
  repeat
    begin
      case GetMode of
        gmCurrent :
          begin
            // ShowMessage ('GetRecord Current');
            if (FCurrentRecord >= FRecordCount+fDataFileHeader.DeletedCount) or
                (FCurrentRecord < 0) then
              Result := grError;
          end;
        gmNext :
          begin
            if (fCurrentRecord < (fRecordCount+fDataFileHeader.DeletedCount)-1) then
              Inc (FCurrentRecord)
            else
              Result := grEOF;
          end;
        gmPrior :
          begin
            if (fCurrentRecord > 0) then
              Dec(fCurrentRecord)
            else
              Result := grBOF;
          end;
      end;
      // fill record data area of buffer
      if Result = grOK then
      begin
        _ReadRecord(Buffer, fCurrentRecord );
        {FStream.Position := FDataFileHeader.StartData +
          FRecordSize * FCurrentRecord;
        FStream.ReadBuffer (Buffer^, FRecordSize);}
        ClearCalcFields(Buffer);
        GetCalcFields(Buffer);
        with PRecInfo(Buffer + FRecordInfoOffset)^ do
        begin
          BookmarkFlag := bfCurrent;
          Bookmark := FCurrentRecord;
        end;
      end
      else
        if (Result = grError) and DoCheck then
          raise eBinaryDataSetError.Create (
            'GetRecord: Invalid record');
      Acceptable := pRecordHeader(Buffer)^.DeletedFlag = 0;
      if (GetMode=gmCurrent) and Not Acceptable then Result := grError;
    end;
  until (Result <> grOK) or Acceptable;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalPost
// III: Write the current data to the file
procedure tBinaryDataSet.InternalPost;
begin
  CheckActive;
  if State = dsEdit then
  begin
    // replace data with new data
    {FStream.Position := FDataFileHeader.StartData + (FRecordSize * FCurrentRecord);
    FStream.WriteBuffer (ActiveBuffer^, FRecordSize);}
    _WriteRecord (ActiveBuffer, fCurrentRecord);
  end
  else
  begin
    // always append
    InternalLast;
    {FStream.Seek (0, soFromEnd);
    FStream.WriteBuffer (ActiveBuffer^, FRecordSize);}
    _AppendRecord(ActiveBuffer);
    Inc (FRecordCount);
  end;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalAddRecord
// III: Add the current data to the file
procedure tBinaryDataSet.InternalAddRecord(
  Buffer: Pointer; Append: Boolean);
begin
  // always append
  InternalLast;
  // add record at the end of the file
  {FStream.Seek (0, soFromEnd);}
  pRecordHeader(ActiveBuffer)^.DeletedFlag := 0;
  _AppendRecord(ActiveBuffer);
  {FStream.WriteBuffer (ActiveBuffer^, FRecordSize);}
  Inc (FRecordCount);
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalDelete
// III: Delete the current record
procedure tBinaryDataSet.InternalDelete;
begin
  CheckActive;
  // not supported in this version
{  raise eBinaryDataSetError.Create (
    'Delete: Operation not supported');}
  pRecordHeader(ActiveBuffer)^.DeletedFlag := fDataFileHeader.LastDeleted;
  _WriteRecord(ActiveBuffer,fCurrentRecord);
  {FStream.Position := FDataFileHeader.StartData + (FRecordSize * FCurrentRecord);
  FStream.WriteBuffer (ActiveBuffer^, FRecordSize);}
  fDataFileHeader.LastDeleted := GetRecNo;
  Inc(fDataFileHeader.DeletedCount);
  Dec(fRecordCount);
  fDataFileHeader.RecordCount := fRecordCount;
  WriteHeader;
  Resync([]);
end;

// ____________________________________________________________________________
// tBinaryDataSet.GetFieldData
// III: Move data from record buffer to field
function tBinaryDataSet.GetFieldData (
  Field: TField; Buffer: Pointer): Boolean;
var
  FieldOffset: Integer;
  Ptr: Pointer;
begin
  Result := False;
  if not IsEmpty and (Field.FieldNo > 0) then
  begin
    FieldOffset := Integer (
      FFieldOffset [Field.FieldNo - 1]);
    Ptr := ActiveBuffer;
//    Inc (Ptr, FieldOffset);
    if Assigned(Buffer) then
      Move (pBufArray(Ptr)^[FieldOffset], Buffer^, Field.DataSize)
    else
      if Field.DataType = ftBoolean then
        Move(pBufArray(Ptr)^[FieldOffset], Buffer, 2)
      else
        ShowMessage ('very bad error in get field data');
    Result := True;
  end;
end;

// ____________________________________________________________________________
// tBinaryDataSet.SetFieldData
// III: Move data from field to record buffer
procedure tBinaryDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  FieldOffset: Integer;
  Ptr: Pointer;
begin
  if Field.FieldNo >= 0 then
  begin
    FieldOffset := Integer (
      FFieldOffset [Field.FieldNo - 1]);
    Ptr := ActiveBuffer;
//    Inc (Ptr, FieldOffset);
    if Assigned (Buffer) then
      Move (Buffer^, pBufArray(Ptr)^[FieldOffset], Field.DataSize)
    else
      if Field.DataType = ftBoolean then
        Move (Buffer, pBufArray(Ptr)^[FieldOffset], 2)
      else
        ShowMessage ('very bad error in setfield data');
    DataEvent (deFieldChange, Longint(Field));
  end;
end;

// ____________________________________________________________________________
// tBinaryDataSet.InternalHandleException
// default exception handling
procedure tBinaryDataSet.InternalHandleException;
begin
  // standard exception handling
  Application.HandleException(Self);
end;

procedure Register;
begin
  RegisterComponents('Data Access', [TBinaryDataSet]);
end;

end.
